module Rede (module Rede) where

import ST


data (Enum p, Ord p) =>  State p = State { state :: p -> Int , range :: (p,p) }
type Trans a p = ([a],State p)

merge :: (Enum p, Ord p) => (State p) -> (State p) -> (Int -> Int -> Int) -> State p
merge u v f = State w (min u' v' , max u'' v'')
     where
       w p      = f (state u p) (state v p)
       (u',u'') = range u 
       (v',v'') = range v

s_span :: (Enum p, Ord p) => (State p) -> [p]
s_span u = let (m,n) = range u in [m..n]

posOn, posOff  :: (Enum p, Ord p) => (State p) -> [p]
posOn  s = filter ((>0)  . (state s)) (s_span s)
posOff s = filter ((==0) . (state s)) (s_span s)

infixl 2 .+ , .-

(.+) , (.-) :: (Enum p, Ord p) => (State p) -> (State p) -> State p
(.+) u v = merge u v (+)
(.-) u v = merge u v (-)

ok :: (Enum p, Ord p) => (State p) -> Bool
ok s = all (\p -> state s p >= 0) (s_span s)

instance (Enum p, Ord p) => Eq (State p) where
     (==) s s' = let z = s .- s' in all ((== 0) . state z) (s_span z)
	
toList :: (Enum p, Ord p) => (State p) ->  [(p,Int)]
toList s = filter ((/=0) . snd) (map (\x -> (x, state s x)) (s_span s))

fromList :: (Enum p, Ord p) => [(p,Int)] -> State p
fromList u = State pp (n,m)  
    where  pp x = case (lookup x u) of { Nothing -> 0 ; Just y -> y }
           (n,_) = head u
           (m,_) = last u
	      

instance (Enum p, Ord p, Show p) => Show (State p)
    where
    showsPrec _  s s' = show (toList s) ++ s'

instance (Enum p, Ord p, Read p) => Read (State p)
    where
    readsPrec n s = [(fromList z,s')] where ((z,s'):_) = readsPrec n s

data (Enum p, Ord p, Enum a) =>
    Net p a = Net { fwrd :: a -> State p, bwrd :: a -> State p, top  :: State p , acts :: [a]  }

t_zero :: (Enum p, Ord p) => ST [] (State p) (State p)
t_zero = delta (return . id)

transition :: (Enum p, Ord p, Enum a) => (Net p a) -> [a] -> ST [] (State p) (State p)
transition n t = delta h
    where  h estado =  guard (ok marcas && ok (top n .- livres)) >> return livres
                  where
                  marcas = let f u a =  u .- (fwrd n a)  in foldl f estado t
		  livres = let f u a =  u .+ (bwrd n a)  in foldl f marcas t

----- transies

path :: (Enum p, Ord p, Enum a) => (Net p a) -> [[a]] -> ST [] (State p) (State p)
path n = foldl f t_zero where f p t = p >> transition n t 

alts :: (Enum p, Ord p, Enum a) => (Net p a) -> [[a]] -> ST [] (State p) (Trans a p)
alts n = foldl f mzero  where f p t = p `mplus` (fmap (\x -> (t,x)) (transition n t)) 

explode :: (Enum p, Ord p, Enum a) => (Net p a) ->  ST [] (State p) (Trans a p)
explode n = alts n (power (acts n))

acessiveis :: (Enum p, Ord p, Enum a) => (Net p a) -> (State p)  -> [(State p)]
acessiveis = cover . explode_
      where explode_ n s = map snd (explode n `on` s)

locks:: (Enum p, Ord p, Enum a, Eq a) => (Net p a) ->  (State p) -> Bool
n `locks` s = (explode n `on`  s) == []

---------------------------------------------------------------------------------------------
-- Auxiliares
---------------------------------------------------------------------------------------------

power :: [a] -> [[a]]
power = foldl f [[]] where f u x = u ++ (map (x:) u)

cover :: (Eq a) => (a -> [a]) -> a -> [a]
cover f x = maybe [] fst (loop iter ([],[x]))
    where  iter (u,[])  = Nothing
	   iter (u,a:v) = let h w y = if a == y || y `elem` u || y `elem` v then w else w ++ [y]
                        in Just (a:u , foldl h v (f a))
           
