-- ----------------------------------------------------------------------------------------------
-- ST.hs
-- JMV
-- ----------------------------------------------------------------------------------------------

module ST (
        module ST,
        module MMonad
   	) where

import MMonad
import Random

newtype (MonadPlus m, Functor m) => ST m s a =  ST { st ::  s -> m (a,s) }
 

-- ----------------------------------------------------------------------------------------------
-- Selectors and constructors
-- ----------------------------------------------------------------------------------------------
infixl 1 `on`, `andthen`

on :: (MonadPlus m, Functor m) => ST m s a -> s -> m a
on = lift fst . st

witness :: (MonadPlus m, Functor m) => (s -> m a) -> ST m s a
witness w = ST (lift2 _f w)  where _f s x = (x,s)

andthen :: (MonadPlus m, Functor m) => ST m s a -> (a -> s -> s) -> ST m s a
andthen (ST p) f =  ST (lift _f p) where _f (x,s) = (x,f x s)

unit :: (MonadPlus m, Functor m) => (a -> s -> s) -> a -> ST m s a
unit f = \x -> return x `andthen` f

delta :: (MonadPlus m, Functor m) => (s -> m s) -> ST m s s
delta f = ST (lift (\x -> (x,x)) f) 

-- --------------------------------------------------------------------------------------------
-- Monadic instantiation
-- --------------------------------------------------------------------------------------------

instance  (MonadPlus m, Functor m) => Functor (ST m s) where
    fmap f (ST p) = ST (lift _f p ) where _f (x,s) = (f x, s)

instance (MonadPlus m, Functor m) => Monad (ST m s) where
   return x       = ST (\s -> return (x,s))
   (ST p) >>= f   = ST (join . lift _f p)
	       where _f (a,s) = st (f a) s

instance (MonadPlus m, Functor m) => MonadPlus (ST m s) where
   mzero                 = ST (\_ -> mzero)
   (ST p) `mplus` (ST q) = ST (\s ->  p s `mplus` q s)

-- ----------------------------------------------------------------------------------------------
-- Ranges
-- ----------------------------------------------------------------------------------------------

class (MonadPlus m, Functor m, Monad r) => Range m r a  where
    pick  :: ST m (r a) a

instance (MonadPlus m, Functor m) => Range m [] a where
    pick = ST (\s -> case s of { [] -> mzero ; (x:xs) -> return (x,xs) })

instance (MonadPlus m, Functor m, Integral a) => Range m Id a  where
    pick = ST  (\s -> case s of { (Id x) -> if x > 0 then return (x,Id (x-1)) else mzero })


-- breath :: (MonadPlus m, Functor m) => (s -> [s] -> [s]) -> ST m [s] s
-- breath t = pick `andthen` f  where f s w = w ++ (t s w)

    
